{
  Word mail merge component.
  Made by Robert van Tilburg, August 23, 2000
  robert@ionosphere.nl

  Check out the homepage for the latest version of this component.
  http://delphi.abydos.nl

  Updates:
  August 30, 2000   - Keith Blows added a major speed improvement to the component, it's now
                      20 times faster than before (by removing Word tables from the
                      datasource and using tabs instead). Thanks Keith!

  September 1, 2000 - Again by Keith Blows: The component will now generate a HTML file with
                      tables, and Word will convert the HTML to a DOC. This is much faster
                      and you can still have fields with tabs and carriage return characters.
                      Tested with over a 1000 records.
                      The previous datasource generation methods are still present, set
                      the DataSource property to mdHTML(default), mdTabs, or mdDirectTable.
                      However, mdHTML is strongly recommended because it is the fastest way.

  October 11, 2000 -  Bug fix: after closing Word you coulnd't merge again.

  March 3, 2001 -  Speed improvement:
                      Appending records is faster
                   Bug fix:
                      Svilen Ivanov solved a bug in the mdDirectTable merge method. Due to
                      column widths of 100, Word would return a "Value out of range" error
                      when you have more than 15 fields. The field width is now set to 10
                      instead of 100 to solve this problem.


}

unit MailMerge;

interface

uses
  Classes,
  Word97,
  Sysutils,
  Dialogs, {for MessageDlg}
  Controls {for ModalResult};

type
  EMailMergeError = class(Exception)
  end;

  TMailMergeDatasource = (mdHTML, mdTabs, mdDirectTable);

  TMailMerge = Class(TComponent)
  private
    FWordApp: TWordApplication;
    FWordDoc: _Document;
    FColumns: TStringList;
    FGrid: array of array of String;
    FRecordNr: Integer;
    FDocumentName: TFileName;
    FStatusMessage: String;
    FOnStatusMessage: TNotifyEvent;
    FMailMergeDatasource: TMailMergeDatasource;
    FHasTabCRLF: Boolean;
    function    GetColumns: String;
    procedure   SetColumns(const Value: String);
    function    GetFieldValue(const FieldName: string): String;
    procedure   SetFieldValue(const FieldName, Value: String);
    function    GetColumnNrByName(const ColumnName: String): Integer;
    procedure   SetDocumentName(const Value: TFileName);
    procedure   SetStatusMessage(const Value: String);
    function    GetWordDoc: _Document;
    function    GetColumnCount: Integer;
    function    GetRecordCount: Integer;
    procedure   CreateWordApp;
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   Append;
    procedure   Clear;
    procedure   Merge;
    property    RecordCount: Integer read GetRecordCount;
    property    ColumnCount: Integer read GetColumnCount;
    property    FieldValues[const FieldName: string]: String read GetFieldValue write SetFieldValue; default;
    property    DocumentName: TFileName read FDocumentName write SetDocumentName;
    property    StatusMessage: String read FStatusMessage;
    property    WordApplication: TWordApplication read FWordApp;
    property    WordDocument: _Document read GetWordDoc;
    property    DataSource: TMailMergeDatasource read FMailMergeDatasource write FMailMergeDatasource default mdHTML;
  published
    property    Columns: String read GetColumns write SetColumns;
    property    OnStatusMessage: TNotifyEvent read FOnStatusMessage write FOnStatusMessage;
  end;

const
  TMailMergeDatasourceNames: array[TMailMergeDatasource] of string = (
    'HTML (Fastest, default)',
    'Word Tabs (Fast, but can''t use fields with CRLF and Tabs)',
    'Word DirectTable (Very slow)'
  );

implementation

{ TMailMerge }

constructor TMailMerge.Create(AOwner: TComponent);
begin
  inherited;
  CreateWordApp;
  FColumns := TStringList.Create;
  DataSource := mdHTML;
  FRecordNr := -1;
  Clear;
end;

destructor TMailMerge.Destroy;
begin
  Clear;
  FColumns.Free;
  inherited;
end;

function TMailMerge.GetColumns: String;
var
  I: Integer;
begin
  Result := '';
  for I := 0 to FColumns.Count - 1 do
  begin
    if Result = '' then
      Result := Result + FColumns[I]
    else
      Result := Result + FColumns[I] + ';';
  end;
end;

procedure TMailMerge.SetColumns(const Value: String);
var
  Start, I: Integer;
begin
  Start := 1;
  for I := 1 to Length(Value) do
  begin
    if Value[I] = ';' then
    begin
      FColumns.Add(Trim(Copy(Value, Start, (I - Start))));
      Start := I + 1;
    end;
  end;
  if Value[Length(Value)] <> ';' then
    FColumns.Add(Trim(Copy(Value, Start, MaxInt)));
end;

function TMailMerge.GetFieldValue(const FieldName: String): String;
var
  I: Integer;
begin
  if FRecordNr = -1 then
    raise EMailMergeError.Create('No current record');
  I := GetColumnNrByName(FieldName);
  Result := FGrid[FRecordNr, I];
end;

procedure TMailMerge.SetFieldValue(const FieldName, Value: String);
var
  I: Integer;
begin
  if FRecordNr = -1 then
    raise EMailMergeError.Create('No current record');

  if not FHasTabCRLF then
  begin
    for I := 1 to Length(Value) do
    begin
      if Value[I] in [#9, #10, #13] then
      begin
        FHasTabCRLF := True;
        Break;
      end;
    end;
  end;

  I := GetColumnNrByName(FieldName);
  FGrid[FRecordNr, I] := Value;
end;

procedure TMailMerge.Clear;
var
  I: Integer;
begin
  for I := 0 to RecordCount - 1 do
    SetLength(FGrid[I], 0);
  SetLength(FGrid, 0);
  FColumns.Clear;
  FRecordNr := -1;
  FDocumentName := '';
  FStatusMessage := 'Idle';
  FHasTabCRLF := False;
end;

function TMailMerge.GetColumnNrByName(const ColumnName: String): Integer;
begin
  for Result := 0 to FColumns.Count - 1 do
    if (AnsiCompareText(FColumns[Result], ColumnName) = 0) then
      Exit;
  Raise EMailMergeError.CreateFmt('Column name does not exist: %s', [ColumnName]);
end;

procedure TMailMerge.Append;
begin
  if FColumns.Count = 0 then
    Raise EMailMergeError.Create('There are no columns defined.');
  if Length(FGrid) = FRecordNr + 1 then
  begin
    // Grow list:
    if Length(FGrid) = 0 then
      SetLength(FGrid, 16)
    else
      SetLength(FGrid, Length(FGrid) + (Length(FGrid) div 4)); // Grow 25%
  end;
  Inc(FRecordNr);
  SetLength(FGrid[FRecordNr], FColumns.Count);
end;

procedure TMailMerge.SetDocumentName(const Value: TFileName);
begin
  if (Length(Value) >= 2) and ((Value[2] <> ':') and (Value[1] + Value[2] <> '\\')) then
    Raise EMailMergeError.CreateFmt('document file "%s" must be a full path and file name', [Value]);
  if not FileExists(Value) then
    Raise EMailMergeError.CreateFmt('Document file not found: %s', [Value]);
  FDocumentName := Value;
end;


// Converts any non-standard characters to HTML valid... not very optmized, but works...
function StringToHTML(const szString : String) : String;
begin
  Result := StringReplace(szString, '&', '&amp;', [rfReplaceAll]);
  Result := StringReplace(Result, '<', '&lt', [rfReplaceAll]);
  Result := StringReplace(Result, '>', '&gt', [rfReplaceAll]);
  Result := StringReplace(Result, '"', '&quot;', [rfReplaceAll]);
  Result := StringReplace(Result, #10#13, '<br>', [rfReplaceAll]);
  Result := StringReplace(Result, #13#10, '<br>', [rfReplaceAll]);
  Result := StringReplace(Result, #10, '<br>', [rfReplaceAll]);
  Result := StringReplace(Result, #13, '<br>', [rfReplaceAll]);
end;


procedure TMailMerge.Merge;
var
  oleVal, oleSep, oleRows, oleCols, oleInitColW,              // KB
  oleFormat, oleBool, DataSourceFile, HTMLSourceFile,  FileFormat,
  ConfirmConversions, AddToRecentFiles, FileName, Pause: OleVariant;
  Doc: _Document;
  X, Y: Integer;
  szColWidth : String;
  MyHTMLDoc  : TextFile;
  T: Table;
  szValue, szTemp: String;
  MyRange : Range;
//  I: OleVariant;
begin
  FWordDoc := nil;
  if (FDocumentName = '') then
    Raise EMailMergeError.Create('Document filename not set.');
  DataSourceFile := IncludeTrailingBackslash(ExtractFilePath(FDocumentName)) + 'MailMergeDatasource.temporary.doc';
  HTMLSourceFile := IncludeTrailingBackslash(ExtractFilePath(FDocumentName)) + 'MailMergeDatasource.temporary.htm';

  WordApplication.Disconnect;
  WordApplication.Connect;

  if (FileExists(DataSourceFile)) and (not DeleteFile(DataSourceFile)) then
    Raise EMailMergeError.CreateFmt('Unable to delete file %s', [DataSourceFile]);

  if Length(FGrid) = 0 then
    Raise EMailMergeError.Create('No records, nothing to do.');

  if DataSource = mdHtml then
  begin
    if (FColumns.Count >= 100) then
      Raise EMailMergeError.CreateFmt('A maximum of 100 fields is permitted. Fields specified= %d',
                                     [FColumns.Count]);
  end;

  if (FHasTabCRLF) and (DataSource = mdTabs) then
  begin
    DataSource := mdHTML;
    SetStatusMessage('Tabbed datasource can''t be used, using HTML datasource instead.');
  end;

  if (DataSource = mdHtml) then // The fastest way, preserves #9 and #13#10 characters:
  begin
    if (FileExists(HTMLSourceFile)) and (not DeleteFile(HTMLSourceFile)) then
      Raise EMailMergeError.CreateFmt('Unable to delete file %s', [HTMLSourceFile]);

    // Create the HTML data source...
    SetStatusMessage('Creating HTML datasource');

    {$IFOPT I+}
    try
    {$ELSE}
      IoResult; // Clear error flag
    {$ENDIF}
      AssignFile(MyHTMLDoc, HTMLSourceFile);
      Rewrite(MyHTMLDoc);

    {$IFOPT I+}
    except on E: Exception do
      begin
        Raise EMailMergeError.CreateFmt('Unable to create file %s', [HTMLSourceFile]);
      end;
    end;
    {$ELSE}
    if IoResult <> 0 then
      Raise EMailMergeError.CreateFmt('Unable to create file %s', [HTMLSourceFile]);
    {$ENDIF}

    // Fill columns on 1st row with column names:
    WriteLn(MyHTMLDoc, '<html><head><meta http-equiv="Content-Language" content="en-za">');
    WriteLn(MyHTMLDoc, '<meta http-equiv="Content-Type" content="text/html; charset=windows-1252">');
    WriteLn(MyHTMLDoc, '<meta name="GENERATOR" content="MailMerge"><meta name="ProgId" content="MailMerge.Document">');
    WriteLn(MyHTMLDoc, '<title>MergeData</title></head><body><table border="0" width="100%"><tr>');

    szColWidth := '"' + IntToStr(100 div FColumns.Count) + '%"';
    SetStatusMessage('Filling datasource with fields');
    for X := 0 to FColumns.Count - 1 do
      WriteLn(MyHTMLDoc, '<td width=' + szColWidth + ' bgcolor="#808080">' +
               StringToHTML(FColumns[X]) + '</td>');
    WriteLn(MyHTMLDoc, '</tr>');

    // Fill cells of table:
    for Y := 0 to RecordCount - 1 do
    begin
      WriteLn(MyHTMLDoc, '<tr>');
      for X := 0 to FColumns.Count - 1 do
        WriteLn(MyHTMLDoc, '<td width=' + szColWidth + '>' +
                StringToHTML(FGrid[Y][X]) + '</td>');
      WriteLn(MyHTMLDoc, '</tr>');
    end;

    WriteLn(MyHTMLDoc, '</table></body></html>');
    CloseFile(MyHTMLDoc);

    SetStatusMessage('Converting data to word document');

    // Open created HTML and save as a data source ...
    AddToRecentFiles := False;
    FileFormat := wdOpenFormatAuto;
    ConfirmConversions := False;
    try
      Doc := FWordapp.Documents.Open(HTMLSourceFile,   // filename
                                     ConfirmConversions, // ConfirmConversions
                                     EmptyParam,       // ReadOnly
                                     AddtoRecentFiles, // AddToRecentFiles
                                     EmptyParam,       // PasswordDocument
                                     EmptyParam,       // PasswordTemplate
                                     EmptyParam,       // Revert
                                     EmptyParam,       // WritePasswordDocument
                                     EmptyParam,       // WritePasswordTemplate
                                     FileFormat);      // Format
    except
      on E: Exception do
      begin
        case MessageDlg('MailMerge: Word failed to open the temporary HTML file (error: "' + E.Message + '"). Would you like to try an alternative way (slower) to complete the mail merge?', mtConfirmation, [mbYes, mbNo], 0) of
          mrYes: begin
            if FHasTabCRLF then
              DataSource := mdDirectTable
            else
              DataSource := mdTabs;
            Merge;
            Exit;
          end;
          mrNo: begin
            Raise;
          end;
        end;
      end;  
    end;
    FileFormat := wdFormatDocument;
    Doc.SaveAs(DataSourceFile, // filename
               FileFormat, // fileformat
               EmptyParam, // Lockcomments
               EmptyParam, // Password
               AddtoRecentFiles, // AddtoRecentFiles
               EmptyParam, // WritePassword
               EmptyParam, // ReadOnlyRecommended
               EmptyParam, // EmbedTrueTypeFonts
               EmptyParam, // SaveNativePictureFormat
               EmptyParam, // SaveFormsData
               EmptyParam ); // SaveAsAOCELetter
  end // End of HTML
  else if (DataSource = mdDirectTable) then // The slow way, but preserves #9 and #13#10 characters
  begin
    SetStatusMessage('Creating Word direct table datasource');
    Doc := FWordapp.Documents.Add(EmptyParam, EmptyParam);

    Doc.Tables.Add(Doc.Application.Selection.Range, RecordCount + 1, FColumns.Count);
    T := Doc.Tables.Item(1);

    T.Rows.Item(1).Cells.Shading.BackgroundPatternColorIndex := wdGray25;
    T.Rows.Item(1).Range.Bold := Integer(True);

    // Fill columns on 1st row with column names:
    SetStatusMessage('Filling datasource with fields');
    for X := 0 to FColumns.Count - 1 do
    begin
      T.Columns.Item(X + 1).SetWidth(10, wdAdjustNone);
      // An error used to occur here in older versions of TMailMerge.
      // The column width is now 10 instead of 100 to fix the bug (see code line above) 
      T.Rows.Item(1).Cells.Item(X + 1).Range.Text := FColumns[X];
    end;

    // Fill cells of table:
    for Y := 0 to RecordCount - 1 do
    begin
      for X := 0 to FColumns.Count - 1 do
      begin
        T.Rows.Item(Y + 2).Cells.Item(X + 1).Range.Text := FGrid[Y][X];
      end;
    end;
  end // End of slow way
  else if (DataSource = mdTabs) then // Create a tabbed datasource and convert to table:
  begin
    SetStatusMessage('Creating Word tabbed datasource');
    Doc := FWordapp.Documents.Add(EmptyParam, EmptyParam);

    // Fill columns on 1st row with column names:
    szValue := '';
    SetStatusMessage('Filling datasource with fields');
    for X := 0 to FColumns.Count - 1 do
    begin
      szValue := szValue + FColumns[X];
      if X<>(FColumns.Count - 1) then
        szValue := szValue + #9 else szValue := szValue + #13#10;
    end;

    // Fill 'cells' of tabbed lines:
    for Y := 0 to RecordCount - 1 do
    begin
      for X := 0 to FColumns.Count - 1 do
      begin
        // Strip any carriage returns and tabs from data...
        szTemp := StringReplace(FGrid[Y][X],#13,'', [rfReplaceAll]);
        szTemp := StringReplace(szTemp, #9, '', [rfReplaceAll]);
        szValue := szValue + szTemp;
        if (X <> (FColumns.Count - 1)) then
          szValue := szValue + #9
        else
          szValue := szValue + #13#10;
      end;
    end;

    SetStatusMessage('Adding data');
    oleVal := 0; MyRange := Doc.Range(oleVal, oleVal);
    MyRange.Text := szValue;
    // Convert to table...

    SetStatusMessage('Converting datasource to table');
    oleSep := wdSeparateByTabs; oleRows := RecordCount - 1 + 1; oleCols := FColumns.Count;
    oleInitColW := 10; oleFormat := wdTableFormatSimple3; oleBool := 1;
    {oleAutoFit := wdAutoFitContent;  oleTableBehave := wdWord9TableBehavior;}
    MyRange.ConvertToTable(oleSep, oleRows, oleCols, oleInitColW, oleFormat, oleBool,
                            oleBool, oleBool, oleBool, oleBool, oleBool, oleBool,
                            oleBool, oleBool);

    T := Doc.Tables.Item(1);
    T.Rows.Item(1).Cells.Shading.BackgroundPatternColorIndex := wdGray25;
    T.Rows.Item(1).Range.Bold := Integer(True);
  end; // End of DataSource = mdTab

  if (DataSource = mdTabs) or (DataSource = mdDirectTable) then
  begin
    // Save the data source:
    SetStatusMessage('Saving datasource');
    FileFormat := wdFormatDocument;
    AddToRecentFiles := False;
    Doc.SaveAs(DataSourceFile, // filename
               FileFormat, // fileformat
               EmptyParam, // Lockcomments
               EmptyParam, // Password
               AddtoRecentFiles, // AddtoRecentFiles
               EmptyParam, // WritePassword
               EmptyParam, // ReadOnlyRecommended
               EmptyParam, // EmbedTrueTypeFonts
               EmptyParam, // SaveNativePictureFormat
               EmptyParam, // SaveFormsData
               EmptyParam  // SaveAsAOCELetter
              );
  end;

  SetStatusMessage('Closing datasource');
  Doc.Close(EmptyParam, EmptyParam, EmptyParam);

  // Load template:
  SetStatusMessage('Opening template');
  FileName := FDocumentName;
  Doc := FWordApp.Documents.Open(FileName,   // FileName
                                 EmptyParam, // ConfirmConversion
                                 EmptyParam, // ReadOnly
                                 EmptyParam, // AddToRecentFiles
                                 EmptyParam, // PasswordDocument
                                 EmptyParam, // PasswordTemplate
                                 EmptyParam, // Revert
                                 EmptyParam, // WritePasswordDocument
                                 EmptyParam, // WritePasswordTemplate
                                 EmptyParam); // Format

  SetStatusMessage('Opening datasource');
  Doc.MailMerge.OpenDataSource(DataSourceFile,
                               FileFormat,
                               EmptyParam,
                               EmptyParam,
                               EmptyParam,
                               EmptyParam,
                               EmptyParam,
                               EmptyParam,
                               EmptyParam,
                               EmptyParam,
                               EmptyParam,
                               EmptyParam,
                               EmptyParam,
                               EmptyParam);

  Doc.MailMerge.Destination := wdSendToNewDocument;
  Pause := False;

  SetStatusMessage('Merging');
  Doc.MailMerge.Execute(Pause);
  FWordDoc := FWordApp.ActiveDocument;

  SetStatusMessage('Closing template');
  Doc.Close(EmptyParam, EmptyParam, EmptyParam);
  DeleteFile(DataSourceFile);
  DeleteFile(HTMLSourceFile);

  SetStatusMessage('Showing Word');
  FWordApp.Visible := True;
  FWordApp.Activate;
  FWordDoc.Activate;
  SetStatusMessage('Done!');

end;

procedure TMailMerge.SetStatusMessage(const Value: String);
begin
  FStatusMessage := Value;
  if Assigned(FOnStatusMessage) then
    FOnStatusMessage(Self);
end;

function TMailMerge.GetWordDoc: _Document;
begin
  Result := FWordDoc;
  if Result = nil then
    Raise EMailMergeError.Create('You can only access the WordDoc interface after a successful call to the Merge method.');
end;

function TMailMerge.GetColumnCount: Integer;
begin
  Result := FColumns.Count;
end;

function TMailMerge.GetRecordCount: Integer;
begin
  Result := FRecordNr + 1;
end;

procedure TMailMerge.CreateWordApp;
begin
  FWordApp := TWordApplication.Create(Self);
  FWordApp.AutoConnect := True;
end;

end.
